#drinc:graphics/rastport.g
#drinc:intuition/miscellaneous.g
#drinc:intuition/screen.g
#drinc:intuition/window.g
#drinc:intuition/intuiMessage.g
#drinc:libraries/dos.g
#crt.g

/*
 * CRTlib - a library of routines useful for writing programs which
 *	    do special I/O to CRT screens.
 */

uint BUFFER_LENGTH = 256;		/* only need a line's worth */

ushort
    UNKNOWN = 253,			/* an unknown position (line or col) */
    PEN_ERASE = 0,			/* A pen for erasing */
    PEN_NORMAL = 1,			/* A pen for unhighlighted drawing */
    PEN_HIGHLIGHT = 2;			/* A pen for highlighted drawing */

*Window_t Window;
*RastPort_t RastPort;

uint
    RandomSeed,
    BufferPos,				/* pos in output buffer */
    TopOffset,				/* offset in pixels of first line */
    LeftOffset, 			/* offset in pixels of first column */
    Width,				/* pixels in text area */
    Height;				/* pixels in text area */

ushort
    CharHeight, 			/* height of a character */
    CharWidth,				/* width of a character */
    CharBaseLine,			/* baseline for text */
    NLines,				/* direct copy of CRT_nLines */
    NCols,				/* direct copy of CRT_nCols */
    CurrLine,				/* current line of cursor */
    CurrCol;				/* current column of cursor */

bool
    AbortEnabled,			/* ^C abort enabled */
    HadError,				/* error message on screen */
    CurrentHighLighted; 		/* currently have highlight pens */

[BUFFER_LENGTH] char Buffer;

/* predeclare local routines - done this way so that this library needs only
   one pass during linking. */

extern
    __header(*char p; *ushort pLine)*char,
    __putCursor()void,
    __unCursor()void,
    __backUp()void,
    __doScroll(ushort topLine, lineCount; short scrollCount),
    __openWindow(*char name; uint length, width)bool;

char
    CONTROL_C = '\(0x03)';	/* ASCII CONTROL-C - signal abort */

/*
 * CRT_NLines - return the number of lines on the CRT.
 */

proc CRT_NLines()ushort:

    NLines
corp;

/*
 * CRT_NCols - return the number of columns on the CRT.
 */

proc CRT_NCols()ushort:

    NCols
corp;

/*
 * CRT_Line - return the screen line the cursor is currently on.
 */

proc CRT_Line()ushort:

    CurrLine
corp;

/*
 * CRT_Column - return the screen column the cursor is currently in.
 */

proc CRT_Column()ushort:

    CurrCol
corp;

/*
 * CRT_AbortEnable - enable the ^C abort.
 */

proc CRT_AbortEnable()void:

    AbortEnabled := true;
corp;

/*
 * CRT_AbortDisable - disable the ^C abort.
 */

proc CRT_AbortDisable()void:

    AbortEnabled := false;
corp;

/*
 * CRT_Chars - a long string is displayed on the Crt screen. If it
 *	       won't all fit at once, it is broken up into segments
 *	       and displayed in segments, with the user pressing
 *	       SPACE to continue. The string is broken on word
 *	       boundaries. Contained \r\n are maintained. Multiple
 *	       contiguous spaces are maintained. On exit, the text
 *	       remains on the screen, and the cursor is at the beginning
 *	       of the last screen line, which is blank.
 */

proc CRT_Chars(register *char s)void:
    register *char s1;
    register ushort line, colsLeft;

    CRT_ClearScreen();
    line := 0;
    colsLeft := NCols;
    while s* ~= '\e' do
	s1 := s;
	if s* = ' ' then
	    while s1* = ' ' do
		s1 := s1 + 1;
	    od;
	    if s1 >= s + NCols then
		s1 := s + (NCols - 1);
	    fi;
	    if make(s1 - s, ushort) >= colsLeft then
		CRT_PutChars("\r\n");
		colsLeft := NCols;
		line := line + 1;
		if line = NLines - 1 then
		    line := 0;
		    CRT_Continue();
		    CRT_ClearScreen();
		fi;
		if s1 - s = 1 then
		    s := s1;
		fi;
	    fi;
	    while s ~= s1 do
		CRT_PutChar(' ');
		s := s + 1;
		colsLeft := colsLeft - 1;
	    od;
	elif s* = '\r' or s* = '\n' then
	    CRT_PutChars("\r\n");
	    colsLeft := NCols;
	    line := line + 1;
	    if line = NLines - 1 then
		line := 0;
		CRT_Continue();
		CRT_ClearScreen();
	    fi;
	    s := s +
		if s* = '\r' and (s + 1)* = '\n' or
			s* = '\n' and (s + 1)* = '\r' then
		    2
		else
		    1
		fi;
	else
	    while s1* ~= '\e' and s1* ~= ' ' and
		    s1* ~= '\r' and s1* ~= '\n' do
		s1 := s1 + 1;
	    od;
	    if s1 >= s + NCols then
		s1 := s + (NCols - 1);
	    fi;
	    if make(s1 - s, ushort) >= colsLeft then
		CRT_PutChars("\r\n");
		colsLeft := NCols;
		line := line + 1;
		if line = NLines - 1 then
		    line := 0;
		    CRT_Continue();
		    CRT_ClearScreen();
		fi;
	    fi;
	    while s ~= s1 do
		CRT_PutChar(s*);
		s := s + 1;
		colsLeft := colsLeft - 1;
	    od;
	fi;
    od;
    if line ~= NLines - 1 then
	CRT_Continue();
    fi;
corp;

/*
 * CRT_Menu - display and allow selection from a menu on the screen.
 *	      The passed chars value contains a header, \e, zero or more
 *	      selections, and a prompt. Each selection consists of a text
 *	      string and will be preceeded on the display by a 1 or 2 digit
 *	      number which the user must type to select that entry. The value
 *	      returned is the 1-origin number of the selected item, or 0 to
 *	      indicate that the user did not wish to pick a selection.
 */

proc CRT_Menu(register *char menu)ushort:
    int BUFFLENGTH = 10;
    register *char p;
    register ushort n, col;
    ushort line, column;
    [BUFFLENGTH] char buffer;

    CRT_ClearScreen();
    menu := __header(menu, &line);
    if line > NLines - 5 then
	line := NLines - 5;
    fi;
    /* first, determine the number and width of the selections so we can
       center them on the screen: */
    p := menu;
    n := 0;
    column := 0;
    while p* ~= '\e' do
	n := n + 1;
	col := 0;
	while p* ~= '\e' do
	    p := p + 1;
	    col := col + 1;
	od;
	if col > column then
	    column := col;
	fi;
	p := p + 1;
    od;
    if n > NLines - line - 2 then
	n := NLines - line - 2;
    fi;
    line := line + (NLines - line - n - 2) / 2;
    if column > NCols - 5 then
	column := NCols - 5;
    fi;
    column := (NCols - column - 5) / 2;
    /* now display the menu: */
    n := 0;
    while menu* ~= '\e' do
	n := n + 1;
	if line < NLines - 1 then
	    CRT_Move(line, column);
	    line := line + 1;
	    CRT_PutChar(if n > 9 then n / 10 + '0' else ' ' fi);
	    CRT_PutChar(n % 10 + '0');
	    CRT_PutChars(" - ");
	    col := column + 5;
	    while menu* ~= '\e' do
		if col < NCols then
		    CRT_PutChar(menu*);
		    col := col + 1;
		fi;
		menu := menu + 1;
	    od;
	else
	    while menu* ~= '\e' do
		menu := menu + 1;
	    od;
	fi;
	menu := menu + 1;
    od;
    menu := menu + 1;
    /* now allow the user to make a selection: */
    while
	CRT_ClearLine(NLines - 2);
	CRT_PutChars(menu);
	CRT_GetLine(&buffer[0], BUFFLENGTH);
	CRT_ClearError();
	p := &buffer[0];
	while p* = ' ' do
	    p := p + 1;
	od;
	line := 0;
	if buffer[0] = '\e' then
	    false
	else
	    while p* >= '0' and p* <= '9' do
		line := line * 10 + (p* - '0');
		p := p + 1;
	    od;
	    if p* ~= '\e' then
		CRT_Error("Bad character; try again.");
		true
	    elif line > n then
		CRT_Error("Selection too big; try again.");
		true
	    else
		false
	    fi
	fi
    do
    od;
    line
corp;

/*
 * __header - local routine to print and count a multi-line header.
 */

proc __header(register *char header; *ushort pLine)*char:
    register *char p;
    register ushort line, col;

    line := 0;
    while header* ~= '\e' do
	col := 0;
	p := header;
	while p* ~= '\e' and p* ~= '\r' and p* ~= '\n' do
	    p := p + 1;
	    col := col + 1;
	od;
	if col > NCols then
	    col := NCols;
	fi;
	CRT_Move(line, (NCols - col) / 2);
	col := 0;
	while header* ~= '\e' and header* ~= '\r' and header* ~= '\n' do
	    if line < NLines and col < NCols then
		col := col + 1;
		CRT_PutChar(header*);
	    fi;
	    header := header + 1;
	od;
	if header* = '\r' then
	    header := header + 1;
	    if header* = '\n' then
		header := header + 1;
	    fi;
	elif header* = '\n' then
	    header := header + 1;
	fi;
	line := line + 1;
    od;
    pLine* := line;
    header + 1
corp;

/*
 * CRT_Error - display an error comment on the last line of the screen.
 */

proc CRT_Error(*char mess)void:

    HadError := true;
    CRT_Move(NLines - 1, 0);
    CRT_PutChars(mess);
corp;

/*
 * CRT_ClearError - clear any error message on the screen.
 */

proc CRT_ClearError()void:

    if HadError then
	HadError := false;
	CRT_ClearLine(NLines - 1);
    fi;
corp;

/*
 * CRT_Continue - issue a message, then wait until user presses space bar.
 *		  Message is centered on the last screen line.
 *		  Last line is cleared after user's reply, and cursor is left
 *		  in bottom left corner of screen.
 */

proc CRT_Continue()void:
    char c;

    CRT_Center(NLines - 1, "Press SPACE to continue");
    while
	c := CRT_ReadChar();
	if c = CONTROL_C and AbortEnabled then
	    CRT_Abort();
	fi;
	c ~= ' '
    do
    od;
    CRT_ClearLine(NLines - 1);
corp;

/*
 * CRT_Ask - ask a question and get a proper yes/no reply.
 *	     Question is asked on the last screen line. After a valid answer
 *	     is received, the last line is cleared and the cursor is left
 *	     in the bottom left corner of the screen.
 */

proc CRT_Ask(*char question)bool:
    register char c;

    while
	CRT_ClearLine(NLines - 1);
	CRT_PutChars(question);
	/* turn cursor on */
	__putCursor();
	c := CRT_ReadChar();
	/* turn cursor back off */
	__unCursor();
	if AbortEnabled and c = CONTROL_C then
	    CRT_Abort();
	fi;
	c ~= 'y' and c ~= 'Y' and c ~= 'n' and c ~= 'N'
    do
    od;
    CRT_ClearLine(NLines - 1);
    c = 'y' or c = 'Y'
corp;

/*
 * CRT_Reply1 - get a string into the passed buffer. All of the
 *		standard CP/M line editing stuff can be used here.
 *		The prompt/reply are on 6th line of empty screen. After the
 *		user's input is received, the screen is cleared and the
 *		cursor is left in the top left hand corner of the screen.
 */

proc CRT_Reply1(*char buffer; ushort buffLen; *char prompt)void:

    while
	CRT_ClearScreen();
	CRT_Move(5, 0);
	CRT_PutChars(prompt);
	CRT_GetLine(buffer, buffLen);
	buffer* = '\e'
    do
    od;
    CRT_ClearScreen();
corp;

/*
 * CRT_Reply2 - this is similar to CRT_Reply1, except that the question is
 *		asked on the last line of the screen and an empty reply
 *		is acceptable.
 */

proc CRT_Reply2(*char buffer; ushort length; *char prompt)void:

    CRT_ClearLine(NLines - 1);
    CRT_PutChars(prompt);
    CRT_GetLine(buffer, length);
    CRT_ClearLine(NLines - 1);
corp;

/*
 * CRT_GetLine - get an input line from the current cursor position.
 */

proc CRT_GetLine(*char buffer; ushort length)void:
    char ch;

    buffer* := '\e';
    pretend(CRT_Read(buffer, length, false, false, "", &ch), void);
corp;

/*
 * CRT_Read - routine to get a reply from the screen, while
 *	      supporting the full CP/M line editing stuff (well, almost).
 */

proc CRT_Read(*char buffer; ushort length; bool skipNext, skipPrev;
		     *char terminators, terminator)bool:
    char CONTROL_X = '\(0x18)';
    register *char buff, p;
    register ushort l;
    register char c;
    bool changed;

    buff := buffer;
    l := length;
    changed := false;
    /* turn cursor on */
    __putCursor();
    while
	if l = 1 and skipNext then
	    c := '\e';
	    false
	else
	    c := CRT_ReadChar();
	    if c = '\r' or c = '\b' and l = length and skipPrev then
		false
	    else
		p := terminators;
		while p* ~= '\e' and p* ~= c do
		    p := p + 1;
		od;
		p* = '\e'
	    fi
	fi
    do
	if AbortEnabled and c = CONTROL_C and l = length then
	    CRT_Abort();
	fi;
	if c = '\b' then
	    if l ~= length then
		l := l + 1;
		__backUp();
		buff := buff - 1;
	    fi;
	elif c = CONTROL_X then
	    while l ~= length do
		l := l + 1;
		__backUp();
	    od;
	    buff := buffer;
	elif c >= ' ' and c <= '~' then
	    if l ~= 1 then
		l := l - 1;
		CRT_PutChar(c);
		buff* := c;
		buff := buff + 1;
		changed := true;
	    fi;
	fi;
    od;
    /* turn cursor off */
    __unCursor();
    if changed then
	buff* := '\e';
    fi;
    terminator* := c;
    changed
corp;

/*
 * backUp - local routine to back up one character on screen.
 */

proc __backUp()void:

    CRT_PutChars("\b \b");
corp;

/*
 * CRT_Reset - reset the terminal by sending terminate and init seqs.
 */

proc CRT_Reset()void:

    CRT_ClearScreen();
corp;

/*
 * CRT_Abort - terminate a run which used CRT I/O.
 */

proc CRT_Abort()void:

    CloseWindow(Window);
    CloseIntuitionLibrary();
    Exit(10);
corp;

/*
 * CRT_Random - return a pseudo random number in the given range.
 */

proc CRT_Random(uint numberRange)uint:

    RandomSeed := RandomSeed * 2113 + 13367;
    RandomSeed := (RandomSeed >> 8) + (RandomSeed << 8);
    RandomSeed % numberRange
corp;

/*
 * CRT_Center - center the given string at the center of the given line.
 */

proc CRT_Center(ushort line; *char mess)void:
    register *char p;
    register ushort len;

    p := mess;
    len := 0;
    while p* ~= '\e' do
	p := p + 1;
	len := len + 1;
    od;
    if len > NCols then
	len := NCols;
    fi;
    CRT_Move(line, (NCols - len) / 2);
    CRT_PutChars(mess);
corp;

/*
 * CRT_ClearTail - clear from the current position to the end of the line.
 */

proc CRT_ClearTail()void:

    SetAPen(RastPort, PEN_ERASE);
    RectFill(RastPort, LeftOffset + CurrCol * CharWidth,
	     TopOffset + CurrLine * CharHeight,
	     LeftOffset + Width - 1,
	     TopOffset + (CurrLine + 1) * CharHeight - 1);
    SetAPen(RastPort, PEN_NORMAL);
    CurrentHighLighted := false;
corp;

/*
 * CRT_ClearToEnd - clear the given line and clear to end-of-screen.
 */

proc CRT_ClearToEnd(ushort line)void:

    if line >= NLines then
	line := NLines - 1;
    fi;
    CurrLine := line;
    CurrCol := 0;
    SetAPen(RastPort, PEN_ERASE);
    RectFill(RastPort, LeftOffset, TopOffset + line * CharHeight,
	     LeftOffset + Width - 1, TopOffset + Height - 1);
    SetAPen(RastPort, PEN_NORMAL);
    CurrentHighLighted := false;
corp;

/*
 * CRT_ClearLine - clear the given screen line (0 at top). The cursor
 *		   is left at the left of the cleared line.
 */

proc CRT_ClearLine(ushort line)void:

    if line >= NLines then
	line := NLines - 1;
    fi;
    CurrLine := line;
    CurrCol := 0;
    SetAPen(RastPort, PEN_ERASE);
    RectFill(RastPort, LeftOffset, TopOffset + line * CharHeight,
	     LeftOffset + Width - 1,
	     TopOffset + CurrLine * CharHeight - 1);
    SetAPen(RastPort, PEN_NORMAL);
    CurrentHighLighted := false;
corp;

/*
 * CRT_ClearLines - clear the given number of lines, starting with the given.
 */

proc CRT_ClearLines(ushort line, count)void:

    if line >= NLines then
	line := NLines - 1;
    fi;
    CurrLine := line;
    CurrCol := 0;
    if count + line > NLines then
	count := NLines - line;
    fi;
    if count ~= 0 then
	SetAPen(RastPort, PEN_ERASE);
	RectFill(RastPort, LeftOffset, TopOffset + line * CharHeight,
		 LeftOffset + Width - 1,
		 TopOffset + (CurrLine + count) * CharHeight - 1);
	SetAPen(RastPort, PEN_NORMAL);
	CurrentHighLighted := false;
    fi;
corp;

/*
 * CRT_InsertLines - insert lines before the given line.
 */

proc CRT_InsertLines(ushort line, count)void:

    __doScroll(line, count, -1);
corp;

/*
 * CRT_DeleteLines - delete lines starting with the given one.
 */

proc CRT_DeleteLines(ushort line, count)void:

    __doScroll(line, count, 1);
corp;

/*
 * CRT_ScrollRegionUp - scroll a region up some lines.
 */

proc CRT_ScrollRegionUp(ushort topLine, lineCount, scrollCount)void:

    __doScroll(topLine, lineCount, scrollCount);
corp;

/*
 * CRT_ScrollRegionDown - scroll a region down some lines.
 */

proc CRT_ScrollRegionDown(ushort topLine, lineCount, scrollCount)void:

    __doScroll(topLine, lineCount, - scrollCount);
corp;

/*
 * CRT_Scroll - scroll the screen up one line. Cursor is left in the first
 *		column of the new empty line at the bottom of the screen.
 */

proc CRT_Scroll()void:

    __doScroll(0, NLines, 1);
    CurrLine := NLines - 1;
corp;

/*
 * __doScroll - general utility for vertical scrolling of regions.
 */

proc __doScroll(ushort topLine, lineCount; short scrollCount)void:

    if topLine >= NLines then
	topLine := NLines - 1;
    fi;
    CurrLine := topLine;
    CurrCol := 0;
    if lineCount + topLine > NLines then
	lineCount := NLines - topLine;
    fi;
    if scrollCount > 0 then
	if scrollCount > lineCount then
	    scrollCount := lineCount;
	fi;
    else
	if scrollCount < -lineCount then
	    scrollCount := -lineCount;
	fi;
    fi;
    if scrollCount ~= 0 then
	ScrollRaster(RastPort, 0, scrollCount * CharHeight, LeftOffset,
		     TopOffset + topLine * CharHeight,
		     LeftOffset + Width - 1,
		     TopOffset + (topLine + lineCount) * CharHeight - 1);
    fi;
corp;

/*
 * CRT_EnterHighLight - enter reverse video (or half-intensity or double-
 *			intensity, or whatever the terminal has)
 */

proc CRT_EnterHighLight()void:

    if not CurrentHighLighted then
	CRT_Flush();
	SetAPen(RastPort, PEN_HIGHLIGHT);
	CurrentHighLighted := true;
    fi;
corp;

/*
 * CRT_ExitHighLight - go back to normal characters
 */

proc CRT_ExitHighLight()void:

    if CurrentHighLighted then
	CRT_Flush();
	SetAPen(RastPort, PEN_NORMAL);
	CurrentHighLighted := false;
    fi;
corp;

/*
 * CRT_DeleteChars - delete characters at the current position.
 */

proc CRT_DeleteChars(ushort n)void:

    if n > NCols - CurrCol then
	n := NCols - CurrCol
    fi;
    if n ~= 0 then
	ScrollRaster(RastPort, n * CharWidth, 0,
		     LeftOffset + CurrCol * CharWidth,
		     TopOffset + CurrLine * CharHeight,
		     LeftOffset + Width - 1,
		     TopOffset + (CurrLine + 1) * CharHeight - 1);
    fi;
corp;

/*
 * CRT_InsertChars - insert blanks at the current position.
 */

proc CRT_InsertChars(ushort n)void:

    if CurrCol + n > NCols then
	n := NCols - CurrCol;
    fi;
    if n ~= 0 then
	ScrollRaster(RastPort, - (n * CharWidth), 0,
		     LeftOffset + CurrCol * CharWidth,
		     TopOffset + CurrLine * CharHeight,
		     LeftOffset + Width - 1,
		     TopOffset + (CurrLine + 1) * CharHeight - 1);
    fi;
corp;

/*
 * CRT_GotChar - return 'true' when a character is ready.
 */

proc CRT_GotChar()bool:

    CRT_Flush();
    WaitForChar(Fd, 1) ~= 0
corp;

/*
 * CRT_GetChar - wait for and read a single character, with cursor enabled.
 */

proc CRT_GetChar()char:
    extern Read(FileHandle_t fd; *char buff; ulong len)ulong;
    char ch;

    /* turn cursor on */
    __putSeq("\(27)[ p");
    CRT_Flush();
    if Read(Fd, &ch, 1) ~= 1 then
	ch := '\e';
    else
	if ch = '\r' then
	    CurrCol := UNKNOWN;
	fi;
    fi;
    /* turn cursor off */
    __putSeq("\(CSI)0 p");
    ch
corp;

/*
 * CRT_ReadChar - wait for and read a single character
 */

proc CRT_ReadChar()char:
    extern Read(FileHandle_t fd; *char buff; ulong len)ulong;
    char ch;

    CRT_Flush();
    if Read(Fd, &ch, 1) ~= 1 then
	ch := '\e';
    else
	if ch = '\r' then
	    CurrCol := UNKNOWN;
	fi;
    fi;
    ch
corp;

/*
 * CRT_PutChars - put a chars string to the terminal
 */

proc CRT_PutChars(register *char p)void:

    while p* ~= '\e' do
	CRT_PutChar(p*);
	p := p + 1;
    od;
corp;

/*
 * CRT_PutChar - routine used to output a character in Crt mode.
 */

proc CRT_PutChar(register char c)void:
    bool forceCursor;
    ushort tl, tc;

    forceCursor := false;
    if
	if c >= ' ' then
	    CurrCol := CurrCol + 1;
	    true
	elif c = '\r' then
	    CurrCol := 0;
	    true
	elif c = '\n' then
	    CurrLine := CurrLine + 1;
	    true
	elif c = '\b' then
	    if CurrCol = 0 then
		forceCursor := true;
		CurrLine :=
		    if CurrLine = 0 then
			NLines - 1
		    else
			CurrLine - 1
		    fi;
		CurrCol := NCols - 1;
		false
	    else
		CurrCol := CurrCol - 1;
		true
	    fi
	elif c = '\t' then
	    while
		CurrCol := CurrCol + 1;
		if CurrLine ~= NLines - 1 or CurrCol ~= NCols then
		    __putChar(' ');
		fi;
		CurrCol ~= NCols and CurrCol & 0x07 ~= 0
	    do
	    od;
	    false
	elif c = '\(0x07)' then 		/* BEL */
	    __putChar('\(0x07)');
	    false
	else
	    false
	fi
    then
	if CurrLine ~= NLines - 1 or CurrCol ~= NCols then
	    __putChar(c);
	fi;
    fi;
    if CurrCol = NCols then
	forceCursor := true;
	CurrCol := 0;
	CurrLine := CurrLine + 1;
    fi;
    if CurrLine = NLines then
	forceCursor := true;
	CurrLine := 0;
    fi;
    if forceCursor then
	tl := CurrLine;
	tc := CurrCol;
	CurrCol := UNKNOWN;
	CRT_Move(tl, tc);
    fi;
corp;

/*
 * CRT_Move - move the cursor to the given line and column on the screen.
 *	      Line 0, column 0 is the top left corner of the screen.
 */

proc CRT_Move(register ushort line, col)void:

    if CurrCol = UNKNOWN then
	CurrLine := UNKNOWN;
    fi;
    if line >= NLines then
	line := NLines - 1;
    fi;
    if col >= NCols then
	col := NCols - 1;
    fi;
    if line ~= CurrLine or col ~= CurrCol then
	CRT_Flush();
	Move(RastPort, LeftOffset + col * CharWidth,
	     TopOffset + line * CharHeight + CharBaseLine);
	CurrLine := line;
	CurrCol := col;
    fi;
corp;

/*
 * CRT__initialize - make Crt point to the proper structure, and do whatever
 *	else is necessary to use direct Crt I/O.
 *	Return 'false' if we cannot open the window.
 */

proc CRT_Initialize(*char title; ushort lines, columns)bool:
    DateStamp_t ds;
    uint length, width;
    bool high, wide;

    if title = nil then
	title := "CRT window";
    fi;
    ignore OpenIntuitionLibrary(0);
    CharWidth := 8;
    CharHeight := 8;
    CharBaseLine := 6;
    if lines <= 1 and columns <= 1 then
	/* special cases - go for largest with and without borders */
	if lines = 0 then
	    /* want max lines, but DO want a title bar */
	    if __openWindow(title, 397, 640) then
		/* interlaced screen */
		NLines := 48;
	    elif __openWindow(title, 197, 640) then
		/* non-interlaced screen */
		NLines := 23;
	    else
		return(false);
	    fi;
	    TopOffset := 11;
	else
	    if __openWindow(title, 400, 640) then
		/* interlaced screen */
		NLines := 50;
	    elif __openWindow(title, 200, 640) then
		/* non-interlaced screen */
		NLines := 25;
	    else
		return(false);
	    fi;
	    TopOffset := 0;
	fi;
	if columns = 0 then
	    NCols := 77;
	else
	    NCols := 80;
	fi;
    else
	if lines < 16 then
	    lines := 16;
	elif lines > 50 then
	    lines := 50;
	fi;
	if columns < 32 then
	    columns := 32;
	elif columns > 80 then
	    columns := 80;
	fi;
	length := make(lines, uint) * CharHeight + 13;
	width := make(columns, uint) * CharWidth + 23;
	if length > 400 then
	    length := 400;
	fi;
	if width > 640 then
	    width := 640;
	fi;
	if __openWindow(title, length, width) then
	    high := length > 200;
	    wide := width > 320;
	    NLines := lines;
	    NCols := columns;
	elif __openWindow(title, 200, width) then
	    high := false;
	    wide := width > 320;
	    NLines := if lines >= 25 then 25 else lines fi;
	    NCols := columns;
	elif __openWindow(title, 200, 320) then
	    high := false;
	    wide := false;
	    NLines := if lines >= 25 then 25 else lines fi;
	    NCols := if columns >= 40 then 40 else columns fi;
	elif __openWindow(title, length, 320) then
	    high := length > 200;
	    wide := false;
	    NLines := lines;
	    NCols := if columns >= 40 then 40 else columns fi;
	else
	    return(false);
	fi;
	if high then
	    if lines = 50 then
		TopOffset := 0;
	    elif lines = 49 then
		TopOffset := 8;
	    else
		TopOffset := 11;
	    fi;
	else
	    if lines >= 25 then
		TopOffset := 0;
	    elif lines = 24 then
		TopOffset := 8;
	    else
		TopOffset := 11;
	    fi;
	fi;
	if wide then
	    if columns = 80 then
		LeftOffset := 0;
	    else
		LeftOffset := 2;
	    fi;
	else
	    if columns >= 40 then
		LeftOffset := 0;
	    else
		LeftOffset := 2;
	    fi;
	fi;
    fi;
    BufferPos := 0;
    CRT_ClearScreen();
    AbortEnabled := true;
    HadError := false;
    DateStamp(&ds);
    RandomSeed := (ds.ds_Days >< ds.dsMinute >< ds.ds_Tick) | 1;
    true
corp;

/*
 * __openWindow - attempt to open a window of the given size.
 */

proc __openWindow(*char title; uint length, width; bool borderless)bool:
    NewWindow_t nw;

    nw := NewWindow_t(
	0, 0, 0, 0, FREEPEN, FREEPEN,
	RAWKEY,
	WINDOWDEPTH | SIMPLE_REFRESH | ACTIVATE,
	nil, nil, nil, nil, nil, 0, 0, 0, 0, WBENCHSCREEN);

    nw.nw_Width := width;
    nw.nw_Height := height;
    nw.nw_Title := title;
    nw.nw_MinWidth := width;
    nw.nw_MinHeight := height;
    nw.nw_MaxWidth := width;
    nw.nw_MaxHeight := height;
    if borderless then
	nw.nw_Flags := WINDOWDEPTH | SIMPLE_REFRESH | BORDERLESS | ACTIVATE;
    fi;
    Window := OpenWindow(&nw);
    Window ~= nil
corp;

/*
 * CRT_ClearScreen - erase the entire screen. Cursor is left in the
 *		     top left hand corner of the screen.
 */

proc CRT_ClearScreen()void:

    SetAPen(RastPort, PEN_ERASE);
    RectFill(RastPort, LeftOffset, TopOffset,
	     LeftOffset + Width - 1, TopOffset + Height - 1)
    CurrLine := 0;
    CurrCol := 0;
    SetAPen(RastPort, PEN_NORMAL);
    CurrentHighLighted := false;
corp;

/*
 * CRT_terminate - the routine called automatically on exit of CRT program.
 */

proc CRT_Terminate()void:

    CRT_Flush();
    CloseWindow(Window);
    CloseIntuitionLibrary();
corp;

/*
 * __putChar - local routine to directly output a character.
 */

proc __putChar(char ch)void:
    extern CRT_Flush()void;

    if BufferPos = BUFFER_LENGTH then
	CRT_Flush();
    fi;
    Buffer[BufferPos] := ch;
    BufferPos := BufferPos + 1;
corp;

/*
 * CRT_Flush - flush our local buffer.
 */

proc CRT_Flush()void:

    Text(RastPort, &Buffer[0], BufferPos);
    BufferPos := 0;
corp;
