#drinc:util.g
#drinc:crt.g

/*
 * FORM.DRC - routines for input/output forms.
 */

/* flags for various options: */

ushort
    FORMHEADERS = 0x01,
    FORMSKIP = 0x02,
    FORMINPUT =0x04,
    FORMOUTPUT = 0x08;

type
    FORMTYPE = enum {FINT, FCHARS},

    FORM = struct {
	*FORM f_next, f_prev;
	FORMTYPE f_type;
	ushort f_line, f_col, f_len, f_headLen;
	*char f_heading;
	*bool f_changed;
	union {
	    struct {
		*int i_new;
		int i_old;
		proc(int n)bool i_check;
	    } i_;
	    struct {
		*char c_old, c_new;
		proc(*char p)bool c_check;
	    } c_;
	} f_;
    };

extern
    __header(*char header; *ushort pLine)*char,
    __fillScreen(*FORM f; bool headers)void,
    __resetCharsValue(*FORM f)void,
    __setCursor(*FORM f)void,
    __putInt(*FORM f)void,
    __putCharsValue(*FORM f)void,
    __freeForm()void;

*FORM Form, Prev;
**FORM FormTail;

/*
 * CRT_FormStart - start the definition of a form.
 */

proc CRT_FormStart()void:

    __freeForm();
    FormTail := &Form;
    Prev := nil;
corp;

/*
 * CRT_FormInt - add an integer element to the current form.
 */

proc CRT_FormInt(ushort l, c, len; *char heading;
			*bool pChanged; *int ptr;
			proc(int n)bool check)void:
    register *FORM f;

    f := new(FORM);
    FormTail* := f;
    FormTail := &f*.f_next;
    f*.f_next := nil;
    f*.f_prev := Prev;
    Prev := f;
    f*.f_type := FINT;
    f*.f_line := l;
    f*.f_col := c;
    f*.f_len := len;
    f*.f_headLen := CharsLen(heading);
    f*.f_heading := heading;
    f*.f_changed := pChanged;
    f*.f_.i_.i_new := ptr;
    f*.f_.i_.i_check := check;
corp;

/*
 * CRT_FormChars - add a string element to the current form.
 */

proc CRT_FormChars(ushort l, c, len; *char heading;
			  *bool pChanged; *char ptr;
			   proc(*char p)bool check)void:
    register *FORM f;

    f := new(FORM);
    FormTail* := f;
    FormTail := &f*.f_next;
    f*.f_next := nil;
    f*.f_prev := Prev;
    Prev := f;
    f*.f_type := FCHARS;
    f*.f_line := l;
    f*.f_col := c;
    f*.f_len := len;
    f*.f_headLen := CharsLen(heading);
    f*.f_heading := heading;
    f*.f_changed := pChanged;
    f*.f_.c_.c_old := pretend(Malloc(len + 1), *char);
    f*.f_.c_.c_new := ptr;
    f*.f_.c_.c_check := check;
corp;

/*
 * CRT_FormIntOK - dummy checker routine which allows anything.
 */

proc CRT_FormIntOK(int n)bool:

    true
corp;

/*
 * CRT_FormCharsOK - dummy checker routine with allows anything.
 */

proc CRT_FormCharsOK(*char p)bool:

    true
corp;

/*
 * CRT_FormRead - read from an input form.
 */

proc CRT_FormRead(*char header; ushort flags; *char terminators)char:
    char
	CONTROL_Q = '\(0x11)',
	CONTROL_R = '\(0x12)',
	CONTROL_Z = '\(0x1a)',
	ESCAPE	  = '\(0x1b)';
    register *FORM f;
    register *char p;
    register int n;
    ushort line;
    char ch;
    [7] char buffer;
    bool skip, changed, isNeg;

    skip := flags & FORMSKIP ~= 0x00;
    if flags & FORMHEADERS ~= 0x00 then
	header := __header(header, &line);
    fi;
    f := Form;
    while f ~= nil do
	if f*.f_type = FINT then
	    f*.f_.i_.i_old := f*.f_.i_.i_new*;
	else
	    CharsCopy(f*.f_.c_.c_old, f*.f_.c_.c_new);
	fi;
	f*.f_changed* := false;
	f := f*.f_next;
    od;
    if flags & FORMOUTPUT ~= 0x00 then
	__fillScreen(Form, flags & FORMHEADERS ~= 0x00);
    fi;
    if flags & FORMINPUT ~= 0x00 then
	f := Form;
	while f ~= nil do
	    __setCursor(f);
	    changed :=
		if f*.f_type = FINT then
		    CRT_Read(
			     &buffer[0],
			     if f*.f_len < 6 then f*.f_len + 1 else 7 fi,
			     skip, true, terminators, &ch
		    )
		else
		    CRT_Read(f*.f_.c_.c_new, f*.f_len + 1, skip, true,
			     terminators, &ch)
		fi;
	    CRT_ClearError();
	    if ch = CONTROL_Z or ch = ESCAPE then
		f := Form;
		while f ~= nil do
		    if f*.f_type = FINT then
			f*.f_.i_.i_new* := f*.f_.i_.i_old;
		    else
			CharsCopy(f*.f_.c_.c_new, f*.f_.c_.c_old);
		    fi;
		    f*.f_changed* := false;
		    f := f*.f_next;
		od;
		__fillScreen(f, false);
		f := Form;
	    else
		if f*.f_type = FINT then
		    if ch = CONTROL_R then
			f*.f_changed* := false;
			f*.f_.i_.i_new* := f*.f_.i_.i_old;
			__setCursor(f);
			__putInt(f);
		    elif changed then
			p := &buffer[0];
			while p* = ' ' do
			    p := p + 1;
			od;
			if p* = '-' then
			    p := p + 1;
			    isNeg := true;
			else
			    isNeg := false;
			fi;
			n := 0;
			while p* >= '0' and p* <= '9' do
			    n := n * 10 - (p* - '0');
			    p := p + 1;
			od;
			if not isNeg then
			    n := -n;
			fi;
			if p* ~= '\e' then
			    CRT_Error("Invalid number, try again.");
			    ch := ' ';
			elif not f*.f_.i_.i_check(n) then
			    ch := ' ';
			else
			    f*.f_changed* := true;
			    f*.f_.i_.i_new* := n;
			fi;
			__setCursor(f);
			__putInt(f);
		    fi;
		else
		    if ch = CONTROL_R then
			__resetCharsValue(f);
		    elif changed then
			f*.f_changed* := true;
			__setCursor(f);
			__putCharsValue(f);
			if not f*.f_.c_.c_check(f*.f_.c_.c_new) then
			    __resetCharsValue(f);
			    ch := ' ';
			fi;
		    fi;
		fi;
	    fi;
	    if ch = '\r' or ch = '\e' then
		ch := '\r';
		f := f*.f_next;
	    elif ch = '\b' then
		if f*.f_prev ~= nil then
		    f := f*.f_prev;
		fi;
	    elif ch ~= ' ' and ch ~= CONTROL_R and ch ~= CONTROL_Z then
		p := terminators;
		while p* ~= '\e' and p* ~= ch do
		    p := p + 1;
		od;
		if p* ~= '\e' then
		    f := nil;
		fi;
	    fi;
	od;
	ch
    else
	'\e'
    fi
corp;

/*
 * __fillScreen - redraw the screen with values and maybe headers.
 */

proc __fillScreen(register *FORM f; bool headers)void:

    while f ~= nil do
	if headers then
	    CRT_Move(f*.f_line, f*.f_col);
	    CRT_PutChars(f*.f_heading);
	else
	    __setCursor(f);
	fi;
	if f*.f_type = FINT then
	    __putInt(f);
	else
	    __putCharsValue(f);
	fi;
	f := f*.f_next;
    od;
corp;

/*
 * __resetCharsValue - reset a chars field to it's old value.
 */

proc __resetCharsValue(register *FORM f)void:

    f*.f_changed* := false;
    CharsCopy(f*.f_.c_.c_new, f*.f_.c_.c_old);
    __setCursor(f);
    __putCharsValue(f);
corp;

/*
 * __setCursor - put the cursor at this element's input area.
 */

proc __setCursor(*FORM f)void:

    CRT_Move(f*.f_line, f*.f_col + f*.f_headLen);
corp;

/*
 * __putInt - local routine to output an integer.
 */

proc __putInt(*FORM f)void:
    [7]char buffer;
    register *char p;
    register int n;
    register ushort len;
    bool isNeg;

    n := f*.f_.i_.i_new*;
    len := f*.f_len;
    p := &buffer[6];
    p* := '\e';
    if n >= 0 then
	n := -n;
	isNeg := false;
    else
	isNeg := true;
    fi;
    while
	len := len - 1;
	p := p - 1;
	p* := -(n % 10) + '0';
	n := n / 10;
	n ~= 0
    do
    od;
    if isNeg then
	len := len - 1;
	p := p - 1;
	p* := '-';
    fi;
    while len > 0 do
	len := len - 1;
	CRT_PutChar(' ');
    od;
    CRT_PutChars(p);
corp;

/*
 * __putCharsValue - put a chars element's current value out.
 */

proc __putCharsValue(*FORM f)void:
    register short i;

    CRT_PutChars(f*.f_.c_.c_new);
    for i from 1 upto f*.f_len - CharsLen(f*.f_.c_.c_new) do
	CRT_PutChar(' ');
    od;
corp;

/*
 * _F_initialize - automatically invoked code to initialize this module.
 */

proc _F_initialize()void:

    Form := nil;
corp;

/*
 * _F_terminate - automatically invoked code to terminate this module.
 */

proc _F_terminate()void:

    __freeForm();
corp;

/*
 * __freeForm - routine to free the storage for the current form.
 */

proc __freeForm()void:
    register *FORM f, fTemp;

    f := Form;
    while f ~= nil do
	if f*.f_type = FCHARS then
	    Mfree(pretend(f*.f_.c_.c_old, *byte), f*.f_len + 1);
	fi;
	fTemp := f;
	f := f*.f_next;
	free(fTemp);
    od;
corp;
